import pyNCS
import pyAex
import os
import pyST
import threading
import time
import numpy as np
#import pyNCS.AerViewer





setup=pyNCS.Setup()
setup.load('zenzero.xml')
setup.apply()

client=pyAex.netMonStimEventsQueue(host='localhost',autostart=0)

class virtualEventsWriter ( threading.Thread ):
    def __init__(self,aer_events,fps):
        self.device_fd=os.open('/tmp/aerfx2_virtual',os.O_RDWR)
        self.aer_data=np.loadtxt(aer_events,'uint32')        #Consider reading line by line for very large files
        self.fTus=np.uint32(1e6/fps)
        threading.Thread.__init__ ( self )
        self.daemon=True

    def put( self, t0, t ):
        '''
        t0 start time (us)
        t end time (us)
        '''
        ev_intval=self.aer_data[(self.aer_data[:,0]>t0)*(self.aer_data[:,0]<t)]
        os.write(self.device_fd,ev_intval)
    def run ( self ):
        t0=time.time()
        while True:
            t0_fus=np.uint32((time.time()-t0)*1e6)
            self.put(t0_fus,t0_fus+self.fTus)
            t_fus=np.uint32((time.time()-t0)*1e6)
            time.sleep(self.fTus*1e-6-(t_fus-t0_fus)*1e-6)

v=virtualEventsWriter('data.txt',40)

v.start()
client.run()

    #viewer=pyNCS.AerViewer.Aer2DViewer(0,'zenzero')    
    






import numpy as np

def clear_sram_memory(sramId = 1, coreId = 0, chipId = 0):
    bits = []
    for neuronId in range(256):
        #we loop over all cores
        dx = 0;
        sx = 0;
        dy = 0;
        sy = 0;
        destcoreId = 0;  #one hot coded    
        bits.append(neuronId << 7 | sramId << 5 | coreId << 15 | 1 << 17 | 1 << 4 | destcoreId << 18 | sy << 27 | dy << 25 | dx << 22 | sx << 24 | coreId << 28 | chipId <<30);
    return np.array(bits).astype('uint64').tostring()

def set_neurons_sram(
        chipId,
        coreId,
        sramId = 1,
        neurons = list(range(256)),
        destcoreId = 0,
        dx = 0,
        sx = 0,
        dy = 0,
        sy = 0,
        ):

    if hasattr(destcoreId, '__len__'):
        destcoreId = sum([2**i for i in destcoreId])

    bits = []
    for neuronId in neurons:
        #we loop over all cores
        bits.append( neuronId << 7 | sramId << 5 | coreId << 15 | 1 << 17 | 1 << 4 | destcoreId << 18 | sy << 27 | dy << 25 | dx << 22 | sx << 24 | coreId << 28 | chipId <<30)
    return np.array(bits).astype('uint64').tostring()

def clear_camId(
        chipId,
        coreId,
        camId,
        neuronId = 0):
    bits = chipId << 30 | 1 << 17 |coreId << 15 | neuronId<<5 ;
    return np.array(bits).astype('uint64').tostring()

def clear_core_cam( chipId,
                    coreId,
                    ):
    bits = []
    for row in range(1024):
        for col in range(16):
            bits.append(chipId << 30 | 1 << 17 |coreId << 15 | row <<5 | col );
    return np.array(bits).astype('uint64').tostring()

def set_neuron_cam(
        chipId,
        camId,
        ei = 1, #excitatory
        fs = 1, #fast
        srcneuronId = 0, #sending neuron
        destneuronId = 0, #receiving neuron
        srccoreId = 0, #sending core (= extratag)
        destcoreId = 0): #receiving core (not 1 hot coded)
    bits = []
    synapse_row = camId;                 # cam ID
    nrn_1 = (destneuronId & 0xf0)>>4
    nrn_2 = destneuronId & 0x0f
    bits .append( chipId << 30 | ei << 29 | fs << 28 | srcneuronId << 20 | srccoreId << 18 | 1 << 17 | destcoreId << 15 | nrn_1 << 11 | camId << 5 | nrn_2 );
    return np.array(bits).astype('uint64').tostring()

def tau1_core_set(chipId,coreId):
    bits = chipId << 30 |\
            1 << 12 |\
            0 << 11 |\
            coreId << 8
    return np.array(bits).astype('uint64').tostring()

def tau2_core_set(chipId,coreId):
    bits = chipId << 30 |\
            1 << 12 |\
            1 << 11 |\
            coreId << 8
    return np.array(bits).astype('uint64').tostring()

def tau2_set(chipId,coreId,neuronId):
    bits = chipId << 30 |\
            1 << 10 |\
            coreId << 8 |\
            neuronId
    return np.array(bits).astype('uint64').tostring()



from pyST import *
import pyNCS
import pyAex
import os
import pyST
import threading
import time
import sys
import numpy as np
from pyAexServer import ServerStarter


setup=pyNCS.Setup(filename='setups/'+sys.argv[1]+'_setuptype.xml')
setup.load('setups/'+sys.argv[1]+'.xml')
setup.apply()
#
if sys.argv[1]=='localhost':
    aexs=ServerStarter.start_server()
client=pyAex.netClient(host=sys.argv[1],autostart=1,fps=1)
#vSeq=pyAex.virtEventSeq('data.txt',25)
#vSeq.start()
b=SpikeList([],[])
c=SpikeList([],[])
for i in range(0,120):
    b[i+.0625]=STCreate.poisson_generator(500000, t_stop=100)
client.stimulate({3:b},isi=False, verbose=True)

client.stop()
ServerStarter.stop_server(aexs)








    






#!/usr/bin/env python
# ############################################################
# python class to control and save data from cAER via tcp
# author  Federico Corradi - federico.corradi@inilabs.com
# ############################################################

import os
import struct
import socket
import threading
import sys
try:
    import Queue as queue
except ImportError:
    import queue
import time
import errno
import numpy as np
from xml.dom import minidom

class caerCommunicationControlServer:
    def __init__(self, host = 'gillygaloo.ss.uci.edu',  port_control = 4040,  inputbuffersize = 8000):

        if sys.platform=="win32":
            self.USE_MSG_WAITALL = False # it doesn't work reliably on Windows even though it's defined
        else:
            self.USE_MSG_WAITALL = False#hasattr(socket, "MSG_WAITALL")

        if sys.version_info<(3, 0):
            self.EMPTY_BYTES=""
        else:
            self.EMPTY_BYTES=bytes([])

        # Note: other interesting errnos are EPERM, ENOBUFS, EMFILE
        # but it seems to me that all these signify an unrecoverable situation.
        # So I didn't include them in de list of retryable errors.
        self.ERRNO_RETRIES=[errno.EINTR, errno.EAGAIN, errno.EWOULDBLOCK, errno.EINPROGRESS]
        if hasattr(errno, "WSAEINTR"):
            self.ERRNO_RETRIES.append(errno.WSAEINTR)
        if hasattr(errno, "WSAEWOULDBLOCK"):
            self.ERRNO_RETRIES.append(errno.WSAEWOULDBLOCK)
        if hasattr(errno, "WSAEINPROGRESS"):
            self.ERRNO_RETRIES.append(errno.WSAEINPROGRESS)

        self.ERRNO_BADF=[errno.EBADF]
        if hasattr(errno, "WSAEBADF"):
            self.ERRNO_BADF.append(errno.WSAEBADF)

        self.ERRNO_ENOTSOCK=[errno.ENOTSOCK]
        if hasattr(errno, "WSAENOTSOCK"):
            self.ERRNO_ENOTSOCK.append(errno.WSAENOTSOCK)
        if not hasattr(socket, "SOL_TCP"):
            socket.SOL_TCP=socket.IPPROTO_TCP

        self.ERRNO_EADDRNOTAVAIL=[errno.EADDRNOTAVAIL]
        if hasattr(errno, "WSAEADDRNOTAVAIL"):
            self.ERRNO_EADDRNOTAVAIL.append(errno.WSAEADDRNOTAVAIL)

        self.ERRNO_EADDRINUSE=[errno.EADDRINUSE]
        if hasattr(errno, "WSAEADDRINUSE"):
            self.ERRNO_EADDRINUSE.append(errno.WSAEADDRINUSE)

        self.V2 = "aedat" # current 32bit file format
        self.V1 = "dat" # old format
        self.EVT_DVS = 0 # DVS event type
        self.EVT_APS = 1 # APS event
        self.port_control = port_control
        self.inputbuffersize = inputbuffersize
        self.host = host 
        self.s_commands = socket.socket()
        self.caer_logging = False
        self.caer_streaming = False
        self.t = None
        self.ts = None
        self.data_stream = "/tmp/data_stream.aedat"

        #caer control standards
        self.header_length = 28
        self.max_cmd_parts = 5
        self.cmd_part_action = 0
        self.cmd_part_node = 1  
        self.cmd_part_key = 2
        self.cmd_part_type = 3
        self.cmd_part_value = 4
        self.NODE_EXISTS = 0
        self.ATTR_EXISTS = 1
        self.GET = 2
        self.PUT = 3
        self.actions = { "node_exists" : [ 11, self.NODE_EXISTS ], "attr_exists" : [11, self.ATTR_EXISTS ],  "get" : [ 3, self.GET ], "put": [ 3, self.PUT ] }
        self.type_action = { 'bool':0, 'byte':1, 'short':2, \
                             'int' :3, 'long':4, 'float':5, \
                             'double':6, 'string':7 }

    def parse_command(self, command):
        '''
          parse string command
            es string: put /1/1-DAVISFX2/'+str(sensor)+'/aps/ Exposure int 10
        '''    
        databuffer = bytearray(4096)
        node_length = 0
        key_length = 0
        action_code = -1
        cmd_parts = command.split()
        if( len(cmd_parts) > self.max_cmd_parts):
            print('Error: command is made up of too many parts')
            return
        else:
            if( cmd_parts[self.cmd_part_action] != None):      # we got come action 
                action_code = self.actions[cmd_parts[0]][1]
                if(action_code == -1):
                    print("Please specify an action to perform as: get/put..")
                    return
                #do action based on action_code                 
                if(action_code == self.NODE_EXISTS):
                    node_length = len(cmd_parts[self.cmd_part_node]) + 1 
                    databuffer[0] = action_code
                    databuffer[1] = 0 #unused
                    databuffer[10:10+node_length] = self.cmd_parts[self.cmd_part_node]
                    databuffer_length = 10 + node_length
                if(action_code == self.PUT):
                    node_length  = len(cmd_parts[self.cmd_part_node]) + 1
                    key_length = len(cmd_parts[self.cmd_part_key]) + 1 
                    value_length = len(cmd_parts[self.cmd_part_value]) + 1                     
                    databuffer[0] = action_code
                    databuffer[1] = self.type_action[cmd_parts[self.cmd_part_type]]
                    databuffer[2:3] = struct.pack('H', 0)
                    databuffer[4:5] = struct.pack('H', node_length)  
                    databuffer[6:7] = struct.pack('H', key_length)
                    databuffer[8:9] = struct.pack('H', value_length)
                    databuffer[10:10+node_length] = str(cmd_parts[self.cmd_part_node]).encode('utf-8')
                    databuffer[10+node_length:10+node_length+key_length] = str(cmd_parts[self.cmd_part_key]).encode('utf-8')
                    databuffer[10+node_length+key_length:10+node_length+key_length+value_length] = str(cmd_parts[self.cmd_part_value]).encode('utf-8')
                    databuffer_length = 10 + node_length + key_length + value_length
                    #raise Exception
                if(action_code == self.GET):
                    node_length  = len(cmd_parts[self.cmd_part_node]) + 1
                    key_length = len(cmd_parts[self.cmd_part_key]) + 1 
                    value_length = 1                     
                    databuffer[0] = action_code
                    databuffer[1] = self.type_action[cmd_parts[self.cmd_part_type]]
                    databuffer[2:3] = struct.pack('H', 0)
                    databuffer[4:5] = struct.pack('H', node_length)  
                    databuffer[6:7] = struct.pack('H', key_length)
                    databuffer[8:9] = struct.pack('H', 0)
                    databuffer[10:10+node_length] = str(cmd_parts[self.cmd_part_node]).encode('utf-8')
                    databuffer[10+node_length:10+node_length+key_length] = str(cmd_parts[self.cmd_part_key]).encode('utf-8')
                    databuffer[10+node_length+key_length:10+node_length+key_length] = str('').encode('utf-8')
                    databuffer_length = 10 + node_length + key_length 
                    #raise Exception

        return databuffer[0:databuffer_length]

    def open_communication_command(self):
        '''
         open jaer UDP communication
        '''
        # create dgram udp socket
        try:
            self.s_commands = socket.socket()
            self.s_commands.connect((self.host, self.port_control))
        except socket.error as msg:
            print(('Failed to create socket %s' % msg))
            sys.exit()

    def get_header(self, data):
        '''
          get header packet
        '''
        eventtype = struct.unpack('H',data[0:2])[0]
        eventsource = struct.unpack('H',data[2:4])[0]
        eventsize = struct.unpack('I',data[4:8])[0]
        eventoffset = struct.unpack('I',data[8:12])[0]
        eventtsoverflow = struct.unpack('I',data[12:16])[0]
        eventcapacity = struct.unpack('I',data[16:20])[0]
        eventnumber = struct.unpack('I',data[20:24])[0]
        eventvalid = struct.unpack('I',data[24:28])[0]
        return [eventtype, eventsource, eventsize, eventoffset, eventtsoverflow, eventcapacity, eventnumber, eventvalid]

    def close_communication_command(self):
        '''
            close tcp communication
        '''
        try:
            self.s_commands.close()
        except socket.error as msg:
            print(('Failed to close socket %s' % msg))
            print((socket.error))
            sys.exit()


    def send_command(self, string):
        '''
            parse input command and send it to the device
            print the answer
                input string - ie. 'put /1/1-DAVISFX2/'+str(sensor)+'/aps/ Exposure int 100'
        '''    
        cmd = self.parse_command(string)    
        self.s_commands.sendall(cmd)
        msg_header = self.s_commands.recv(4)
        msg_packet = self.s_commands.recv(struct.unpack('H', msg_header[2:4])[0])
        action = struct.unpack('B',msg_header[0:1])[0]
        second = struct.unpack('B',msg_header[1:2])[0]
        print((string+' action='+str(action)+' type='+str(second)+' message='+msg_packet.decode()))
        return msg_packet

    def load_biases(self, xml_file = 'cameras/davis240c.xml', dvs128xml = False):
        '''
            load default biases as defined in the xml file
        '''
        if(dvs128xml):
            xmldoc = minidom.parse(xml_file)
            nodes = xmldoc.getElementsByTagName('node')
            value_zero = nodes[1].attributes['name'].value
            value_cam = nodes[2].attributes['name'].value 
            command = []
            #loop over xml file, get bias values and load them
            for i in range(len(nodes)):
                if(nodes[i].attributes['name'].value == 'bias' ):
                    attrs = dict(list(nodes[i].attributes.items()))
                    base_aa = 'put'
                    bias_node = nodes[i]
                    biases = bias_node.childNodes
                    for j in range(len(biases)):
                        if(biases[j].hasChildNodes()):
                            base_a = dict(list(biases[j].attributes.items()))
                            base_ab = str(attrs['path'])
                            base_b =  base_a
                            #print base_b
                            base_ac = (base_b['key'])
                            base_ad = (base_b['type'])
                            final_v = str(biases[j].firstChild.data)
                            #print final_v
                            cear_command = base_aa + " " + base_ab + " " + base_ac + " " + base_ad + " " + final_v
                            #print cear_command
                            self.send_command(cear_command)                  
        else:
            xmldoc = minidom.parse(xml_file)
            nodes = xmldoc.getElementsByTagName('node')
            value_zero = nodes[1].attributes['name'].value
            value_cam = nodes[2].attributes['name'].value 
            command = []
            #loop over xml file, get bias values and load them
            for i in range(len(nodes)):
                if(nodes[i].attributes['name'].value == 'bias' ):
                    attrs = dict(list(nodes[i].attributes.items()))
                    base_aa = 'put'
                    bias_node = nodes[i]
                    biases = bias_node.childNodes
                    for j in range(len(biases)):
                        if(biases[j].hasChildNodes()):
                            base_a = dict(list(biases[j].attributes.items()))
                            #print base_a
                            base_ab = (base_a['path'])
                            if(biases[j].hasChildNodes()):
                                bias_values = biases[j].childNodes
                                for k in range(len(bias_values)):
                                    if(bias_values[k].hasChildNodes()):
                                        base_b = dict(list(bias_values[k].attributes.items())) 
                                        #print base_b
                                        base_ac = (base_b['key'])
                                        base_ad = (base_b['type'])
                                        final_v = str(bias_values[k].firstChild.data)
                                        #print final_v
                                        cear_command = base_aa + " " + base_ab + " " + base_ac + " " + base_ad + " " + final_v
                                        #print cear_command
                                        self.send_command(cear_command)                  

if __name__ == '__main__':
    c = caerCommunicationControlServer()



from pyNCSre.api.ComAPI import *
import pyCAER.client as caerclient
from pyCAER.utils import doc_inherit

#Has a single client.
#Instanciated object can be reused


class Communicator(ContinuousCommunicatorBase):
    '''
    This class implements the BatchCommunicator API defined by the NCS tools.
    It uses the "AEX net client" to connect of the AEX server of the target host, using TCP.

    Inputs:
    *host:* the hostname of the computer where the AEX board is attached (default: localhost). If localhost is chosen, then ssh commands are *not* used.
    for additional arguments, refer to caerclient.netClient.

    See also:
    `netClient <#pyCaer.caerclient.netClient>`_

    Usage:
    >>> c = Communicator(host = 'localhost', devnum = 0)
    >>> c.run()
    '''
    def __init__(self, *netclient_args, **netclient_kwargs):
        self.args = netclient_args
        self.kwargs = netclient_kwargs
        self.kwargs['autostart'] = False
        self.out = caerclient.pyST.events()
        self._client = None
        ContinuousCommunicatorBase.__init__(self)

    @property
    def client(self):
        if self._client != None:
            return self._client
        else:
            return None

    @client.setter
    def client(self, value):
        if self._client != None and self.client.is_alive():
            self.close()
        self._client = value

    def run(self, stimulus=None, duration=1000, context_manager=None, **kwargs):    
        self.open()
        self.stim(stimulus, duration, context_manager, **kwargs)
        self.close()
        return self.out

    def mon(self, duration):
        return self.client.fetch()

    def stim(self, stimulus, duration=None, context_manager=None, **stim_kwargs):
        stim_kwargs.update([
            ['tDuration', duration],
            ['context', context_manager]])
        self.out = self.client.stimulate(stimulus, **stim_kwargs)
        return self.out

    def open(self):
        self._isopen = True
        self.client = caerclient.AEDATClient(*self.args, **self.kwargs)
        self.client.start()

    def close(self):
        self._isopen = False
        self.client.stop()



from pyNCSre.api.ConfAPI import MappingsBase
from contextlib import contextmanager
from pyCAER.utils import doc_inherit, getuser, flatten 
from pyCAER.maputils import * 
from pyCAER.api.conf_tcpclient import Configurator
import numpy as np
import socket


class Mappings(MappingsBase):

    @doc_inherit
    def __init__(self, host='128.200.83.67', concatenated=True, local_directory='/tmp/', remote_directory='/var/opt/pyncs/', debug=True, *args, **kwargs):
        self.host = host
        self.debug = debug
        self.cur_mappings = []
        self.remote_directory = remote_directory
        self.local_directory = local_directory 
        self.filename = 'mapping_table_' + getuser()
        self.local_filename = self.local_directory+self.filename
        if concatenated:
            self.remote_filename = self.remote_directory + 'mapping_table'
        else:
            self.remote_filename = self.remote_directory + self.filename
        
        #self.clear_mappings()
        super(self.__class__, self).__init__()

    def register_neurosetup(self, neurosetup):
        '''
        Provides a link to the Neurosetup. This is useful for complex parameter
        configuration protocols requiring the sequencing and monitoring of
        address-events
        '''
        self._neurosetup_registered = True
        self._neurosetup = neurosetup
        print("TODO: Checking chip configurator")
        print(type(neurosetup.chips['U0'].configurator))

    @doc_inherit
    def add_mappings(self, mappings):

        n_connections = len(mappings)
        if n_connections>0:
            mappings = np.array(mappings, dtype='uint32')
            nsetup = self.neurosetup
            mon_chad = nsetup.mon.addrPhysicalExtract(mappings[:,0])
            seq_chad = nsetup.seq.addrPhysicalExtract(mappings[:,1])

            #channels = [i for i,c in enumerate(mon_chad) if c is not None ] 
            #assert len(channels) == 1, 'cross chip connections not yet supported'
            
            src_chip = nsetup.mon.extract_channels(mappings[:,0])
            dst_chip = nsetup.seq.extract_channels(mappings[:,1])

            with open(self.local_filename, 'w') as f:
                f.write('#Writing pyNCS Connections, N: {0}\n'.format(n_connections))
                for i in range(n_connections):
                    src_chip, dst_chip = nsetup.mon.extract_channels(mappings[i])
                    src_neuron, src_core = nsetup.mon[src_chip].addrPhysicalExtract(mappings[i:i+1,0])
                    fs, ei, dst_neuron, dst_core = nsetup.seq[dst_chip].addrPhysicalExtract(mappings[i:i+1,1])
                    synapse = fs+(ei*2)
                    s = 'U{:02d}-C{:02d}-N{:03d}->{:01d}-1-U{:02d}-C{:02d}-N{:03d}\n'.format( 
                        src_chip,
                        src_core[0],
                        src_neuron[0],
                        synapse[0], 
                        dst_chip,
                        dst_core[0], 
                        dst_neuron[0])
                    f.write(s)
        else:
            with open(self.local_filename, 'w') as f:
                f.write('#Writing pyNCS Connections, N: {0}\n'.format(n_connections))
            
        
        self._commit()
        self.cur_mappings = mappings
        return None



    @doc_inherit
    def set_mappings(self, mappings):
        self.clear_mappings()
        self.add_mappings(mappings)

#       dst_cores_1hot = sram_mappings[chipId, pre_srccore[conn2make],pre_addr[conn2make],1,:4]
#       dst_cores = np.sum([dst_cores_1hot[:,i]*2**i for i in range(4)], axis=0)

#       for d in np.unique(pre_srccore):
#           self.clear_sram_chip_core(chipId, d)

#       for d in np.unique(post_dstcore):
#           self.clear_cam_chip_core(chipId, d)


    def open(self):
        '''
        Open MapClient
        '''
        self.client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.client.connect((self.host,self.port))
        super(self.__class__, self).open()

    def close(self):
        '''
        Close MapClient
        '''
        self.client.close()
        del self.client
        super(self.__class__, self).close()

    @doc_inherit
    def get_mappings(self):
        return self.cur_table

    @doc_inherit
    def clear_mappings(self):
        self.cur_mappings = []
        return None

    @doc_inherit
    def clear_cam_chip_core(self, chipId, coreId):
        '''
        clear_cam_chip_core(self, chipId, coreId)
        Write zeros to all CAMs (deletes all connection in a core)
        '''
        #data =  clear_core_cam(chipId=chipId, coreId=coreId)
        print(('Clearing CAM on ChipID: {0} CoreID {1}'.format(chipId,coreId)))
        #self.commit(data)

        return None

    @doc_inherit
    def clear_sram_chip_core(self, chipId, coreId):
        #data = []
        #data.append(clear_sram_memory(chipId=chipId, sramId=1, coreId=coreId))
        #print(('Clearing SRAM on ChipID: {0} CoreID {1}'.format(chipId,coreId)))

        #self.commit(data)
        return None

    @contextmanager
    def _commit(self):
        '''
        '''
        import time, ftplib
        conf = None
        for chip in self.neurosetup.chips.values():
            if type(chip.configurator) == Configurator: 
                conf = chip.configurator 
        if conf is None:
            raise Exception('No Dynapse Chip Found\n')
        else:
            ftp = ftplib.FTP(self.host, user='pyncs')
            ftp.storlines('STOR {0}'.format(self.filename),
                    open('{0}'.format(self.local_filename),'rb'))
            ftp.close()
            while conf.get_caer_sshs("/netparser/", "ProgramNetworkFrom.txt", 'bool').strip('\x00'.encode()) is "true":
                print(conf.get_caer_sshs("/netparser/", "ProgramNetworkFrom.txt", 'bool'))
                print("Waiting other processes to finish weight programming")
                time.sleep(.5)
            conf.set_caer_sshs("/netparser/", "net_txt_file", "string", self.remote_filename)
            time.sleep(.3)
            conf.set_caer_sshs("/netparser/", "ProgramNetworkFrom.txt", 'bool', "true")
            while conf.get_caer_sshs("/netparser/", "ProgramNetworkFrom.txt", 'bool').strip('\x00'.encode()) is "true":
                print("Waiting weight programming to complete")
                time.sleep(.5)
        
        



# -*- coding: utf-8 -*-
from setuptools import setup

setup(name='pyCAER',
        version='20180216_r1',
	description='python Caer client and NCS api',
	author='Emre Neftci',
	author_email='eneftci@uci.edu',
	url='https://github.com/nmi-lab/pyCAERdynapse',
	packages=['pyCAER','pyCAER.api'],
        package_dir={'' : 'src'},
        install_requires=['numpy'],
     )






import threading
import socket, struct
try:
    import Queue as queue
except ImportError:
    import queue
#For probing network buffer
import fcntl
from termios import FIONREAD
from contextlib import contextmanager
#AER related
import pyNCSre.pyST as pyST
from .utils import *

#Special events used to control reset and synchonization
GOEVENT = np.array([-1, 0], dtype='uint32').tostring() #for stim = None, this ensures that the hold is released.
HOLDEVENT = np.array([-2, 0], dtype='uint32').tostring() #Hold and reset: the aex device file is opened immediately before the next event is receied

VALID_SHIFT = 0
VALID_WIDTH = 1
VALID_MASK = 2**VALID_WIDTH-1
CHIP_SHIFT = 6
CHIP_WIDTH = 4
CHIP_MASK = 2**CHIP_WIDTH-1
CHIP_NMASK = 2**32-1 - ((2**CHIP_WIDTH-1)<<CHIP_SHIFT)


class AEDATClientBase(threading.Thread):
    '''
    This class starts a Monitor client for the AEDAT server. Once created, this continously reads the monitored events.
    '''
    #Singleton: make sure a client is only instantiated once in a session
    def __init__(self,
            host='gillygaloo.ss.uci.edu',
            port=7778,
            autostart=True,
            qsize=4096,
            ):
        '''
        *MonChannelAddress:* Monitor Channel Addressing object, if omitted, default channel is taken
        *host:* Monitoring AEDAT server hostname.
        *port:* Port of the Monitoring Server (Default 50001)
        *qsize:* The maxsize of the queue
        '''
        threading.Thread.__init__(self)
        self.daemon = True  # The entire Python program exits when no alive non-daemon threads are left
        print(("Connecting to " + host))
        #self.aexfd = os.open("/dev/aerfx2_0", os.O_RDWR | os.O_NONBLOCK)
        self.finished = threading.Event()
        self.buffer = queue.Queue(qsize)  # 8192 packets in Queue=~5min

        self.MonChannelAddress = pyST.getDefaultMonChannelAddress()


        self.recvlock = threading.Lock()
        self._nbufferempty = False
        self.eventsPacket = np.array([])
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.host = host
        self.port_mon = int(port)

        self.sock.connect((self.host, self.port_mon))

        self.data_header = self.sock.recv(20)
        # Set the socket to non-blocking (with timeout) to avoid the thread
        # being stuck. Because of this, I must catch EWOULDBLOCK error
        #Frame period
        self.fT = 1./50
        self.sock.settimeout(5.0)


        #Build decoder functions
        if autostart:
            self.start()

    def run(self):
        while self.finished.isSet() != True:
            t0 = time.time()

            with self.recvlock:
                self.fetch_raw()

            #Fastest frame rate should be 1/fT. Wait in case we've been faster
            t_left = self.fT - (time.time() - t0)
            if t_left > 0:
                time.sleep(t_left)
            # Ok now check how much data is available and choose a multiple of
            # 8 (AE packet size

    def flush(self, verbose=False):
        '''
        empties the buffers.
        '''

        #empty Queue buffer
        with self.recvlock:
            self.buffer.queue.clear()
#           while True:
#               try:
#                   self.buffer.get(False)
#               except queue.Empty:
#                   break

    def stop(self):
        self.finished.set()
        with self.recvlock:
            self.sock.close()

    def __del__(self):
        self.stop()


##################
##Mapper functions
class AEDATMonClient(AEDATClientBase):

    def _recv_packet(self):
        data_ev_head = self.sock.recv(28)  # we read the header of the packet
        while len(data_ev_head)<28:
            data_ev_head += self.sock.recv(28-len(data_ev_head))
            print('reread')


        
        # read header
        eventtype = struct.unpack('H', data_ev_head[0:2])[0]
        eventsource = struct.unpack('H', data_ev_head[2:4])[0]
        eventsize = struct.unpack('I', data_ev_head[4:8])[0]
        eventoffset = struct.unpack('I', data_ev_head[8:12])[0]
        eventtsoverflow = struct.unpack('I', data_ev_head[12:16])[0]
        eventcapacity = struct.unpack('I', data_ev_head[16:20])[0]
        eventnumber = struct.unpack('I', data_ev_head[20:24])[0]
        eventvalid = struct.unpack('I', data_ev_head[24:28])[0]
        next_read = eventcapacity * eventsize
        if(eventtype == 12):
            data = ''.encode()
            while len(data)<next_read:
                data += self.sock.recv(next_read-len(data))
            return data, eventvalid
        else:
            return '', 0

    def put_buffer(self, ev):
        try:
            self.buffer.put(ev, block=False)
        #Drop last and put new
        except queue.Full:
            self.buffer.get(block=True)
            self.buffer.put(ev, block=True)

    @contextmanager
    def _isbuffernotempty(self, ev, nev):
        try:
            yield
            self._nbufferempty = True
        except ValueError:
            if not ev == '' and nev == -1:
                self._nbufferempty = True  # queue is not empty
                pass
            else:
                self._nbufferrempty = False  # queue is not empty
                pass

    def _extract_events(self, ev):
        if ev == '':
            raise ValueError()

        tmp_array = np.fromstring(ev, dtype='uint32')

        #Do swap if big endian system
        if sys.byteorder == 'big':
            tmp_array.byteswap(True)
        evs = events(tmp_array.reshape(-1, 2), atype='p')

        return evs

    def fetch_raw(self):
        '''
        This functions is called by threading.Thread.run to read from the TCP fifo and add an event packet (a STas.events object) to the client's buffer in physical format.
        '''
        eventsPacket, nev = self._recv_packet()
        
        with self._isbuffernotempty(eventsPacket, nev):
            #Read from binary
            evs = self._extract_events(eventsPacket)
            #extract logical for all channels
            self.put_buffer(evs)

        return self._nbufferempty

    def fetch(self):
        '''
        Empties the buffer and returns its contents
        returns numpy.array in raw address - timestamp format.
        '''
        x = events(atype='p')
        while True:
            try:
                x.add_adtmev(self.buffer.get(block=False).get_adtmev())
            except queue.Empty:
                return x.get_adtmev()

    def listen(self, tDuration=1000, filter_duplicates=False):
        '''
        empties the queue & returns a SpikeList containing the data that was in the Queue
        synposis: client.flush(); time.sleep(1); out=client.listen()
        @author emre@ini.phys.ethz.ch, andstein@student.ethz.ch

        *tDuration*: is a float defining the duration to be listened (in ms)
        *output*: is a string defining in what form the output should be returned. Default is SpikeList where a NeuroTools.signal SpikeList is returned. Other possiblities are 'array' where a normalized numpy array is returned and 'raw' where a pyST.channelEvents object is returned.
        *filter_duplicates* if True erase in all channels double events within the a 0.01ms time-frame (buggy hw)
        '''

        #initialize output channelEvents
        out = events(atype='p')
        target_delta = tDuration*1000
        data_delta = 0
        while data_delta < target_delta:
            try:
                evs = self.buffer.get(block=True)
            except queue.Empty:
                continue

            if evs.get_nev() > 0:
                ad = evs.get_ad()
                tm = evs.get_tm()
                valid = (ad&VALID_MASK).astype('bool')
                chip = (ad[valid]>>CHIP_SHIFT)&CHIP_MASK
                ad[valid] = ad[valid]&CHIP_NMASK
                ad = ad[valid]>>VALID_WIDTH
                tm = tm[valid]
                ad += chip<<28
                out.add_adtm(ad,tm)
            #Assumes isi
            data_delta = (out.get_tm()[-1] - out.get_tm()[0])
                    
        tm = out.get_tm()
        tm = tm - tm[0]
        out.set_tm(tm)
        return out.get_adtmev()


class AEDATClient(AEDATMonClient):
    '''
    This class starts a Monitor/Stimulation client for the AEDAT server. Once started, this continously reads the monitored events.
    '''
    def __init__(self,
            host='gillygaloo.ss.uci.edu',
            host_stim=None,
            port_mon=7778,
            port_stim=50002,
            autostart=True,
            qsize=4096,
            ):
        '''
        *MonChannelAddress:* Monitor Channel Addressing object, if omitted, default channel is taken
        *SeqChannelAddress:* Sequencer Channel Addressing object, if omitted, default channel is taken
        *host:* Monitoring and Sequencing AEDAT server hostname. (Must be same)
        *port_mon:* Port of the Monitoring Server (Default 50001)
        *port_stim:* Port of the Monitoring Server (Default 50002)
        *qsize*: is the size of the queue (FIFO). This is automatically adjusted if the buffer is too small.
        '''

        self.MonChannelAddress = pyST.getDefaultMonChannelAddress()
        self.SeqChannelAddress = pyST.getDefaultSeqChannelAddress()

        AEDATMonClient.__init__(
                self, 
                host=host,
                port=port_mon,
                autostart=False,
                qsize=qsize,
                )

        if host_stim == None:
            self.host_stim = host
        else:
            self.host_stim = host_stim

        #self.aexfd = os.open("/dev/aerfx2_0", os.O_RDWR | os.O_NONBLOCK)
        #self.stim_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        #self.stim_sock.connect((self.host_stim, int(port_stim)))

        if autostart:
            self.start()

    def stimulate(self,
            stim=None,
            tDuration=None,
            context=None,
            filter_duplicates=False,
            debug=False,
            verbose=True,
            ):
        '''
        NOTE: OLD DOCUMENTATION !!
        This function stimulates through AEDAT server. **NOT** network lag invariant!
        *stim:* should be a SpikeList (usable by channelAddressing.exportAER())
        *tDuration:* in ms (Default: guess from stim or 1000 if stim==None).
        *isi:* Whether the input should be send to server as inter-spike intervals.
        *send_reset_event* resets the AEDAT board when stimulated (experimental).
        *output* format of output, spikelist/array/raw/
        *context* is a context which is executed before and after the stimulation (for sync functions and clean-up functions for example, see the python contextlib module).
        *filter_duplicates* if True erase in all channels double events within the a 0.01ms time-frame (buggy hw)
        '''

        if context == None:
            context = empty_context

        stim = events(None, 'p') #stimulus disabled!

        if sys.byteorder == 'big':
            stimByteStream = stim.get_tmadev().byteswap().tostring()
        else:
            stimByteStream = stim.get_tmadev().tostring()

        if verbose:
            print("Bytecoding input...")
            #print((np.fromstring(stimByteStream, 'uint32')))
            print("Done")

        if tDuration == None:
            tDuration = np.sum(stim.get_tm()) * 1e-3  # from us to ms

        if self.buffer.maxsize / self.fT < tDuration:
            while self.buffer.maxsize / self.fT < tDuration:
                self.buffer.maxsize *= 2

        #Clean up pre-stimulus data
        if verbose:
            print("Flushing pre-stimulus data...")
        self.flush()

        if verbose:
            print(("Sending to " + self.host))

        if len(stim)>0:
            if verbose:
                print((np.fromstring(stimByteStream, 'uint32')))

            if send_reset_event:
                self.stim_sock.send(HOLDEVENT)
                self.flush()  # Be sure that there are no events left in the buffer
                # Since the monitor is held, no events can come in until the
                # reset. Well... in theory
                #Bug: Somehow, a few event still get through...
            with context():
                #self.buffer.put(-1, block=True)
                self.stim_sock.send(stimByteStream)
                time.sleep(tDuration * 1e-3 + tDuration * 1.0 * 1e-3 + .100)

        else:
            if verbose:
                print(("Waiting " + str(
                    tDuration) + "ms " + "for stimulation to finish"))

        #####
        # Throw away pre-stimulus data
        #if verbose:
        #    print "Throwing away pre-stimulus-data"

        #while self.buffer.get() != -1:
        #    pass

        if verbose:
            print("Recieving...")
        received = self.listen(tDuration=tDuration, filter_duplicates=filter_duplicates)
        return received
            
    def stop(self):
        self.finished.set()
        with self.recvlock:  # Be sure that there is no fetching going on
            self.sock.close()
        #with self.stimlock # there is no stimlock
        #self.stim_sock.close()

    def __del__(self):
        self.stop()

            
if __name__ == '__main__':
    c = AEDATMonClient()





# api file for pyNCS conf
# import base classes from pyNCS
from pyNCSre.api.ConfAPI import ConfiguratorBase
from pyNCSre import doc_inherit #for inheriting documentation from base classes
from pyCAER import caer_communication
from lxml import etree

class Parameter:
    def __init__(self, name, path):
        '''
        Parameter(parameters, configurator)
        params_dict: dictionary of parameters and values
        This object is designed to be used with the configurator to set parameters
        '''
        self.path = path
        self.SignalName = name
        self.attr = {}

    def add_attr(self, masterkey, attrdict, value = None):
        self.attr[masterkey] = ParameterAttribute(attrdict, value)

    def __str__(self):
        return str([str(s)+' ' for s in self.attr])

    def __repr__(self):
        return str([str(s)+' ' for s in self.attr])

class ParameterAttribute:
    def __init__(self, attrdict, value):
        self.data = attrdict
        self.value = value

    def __str__(self):
        return str(self.data)



class Configurator(ConfiguratorBase):
    def __init__(self, host='localhost', port=4040):
        '''
        Configuration class to set biases through the AMDA boards designed at the Institute of Neuroinformatics.
        Input:
        *host:* hostname of the computer where the AMDA board is attached
        '''
        ConfiguratorBase.__init__(self)
        self.host = str(host)
        self.port = int(port)
        self.client = caer_communication.caerCommunicationControlServer(host=self.host, port_control=self.port)
        self.open()

    def load_parameter_definitions(self, inp, chip):
        '''
        Parse xml file or element tree to generate the object
        '''
        if isinstance(inp, str):
            # parse the file
            doc = etree.parse(inp)
        else:
            doc = inp
        alld = dict()
        doc = doc.find('.//node[@name="{0}"]'.format(chip.id.upper()))
        for biases_root in doc.findall('.//node[@name="bias"]'):
            for br in biases_root.iterfind('.//node'):
                new_parameter = Parameter(br.attrib['name'], br.attrib['path'])
                for attr in br.iterfind('.//attr'):
                    new_parameter.add_attr(attr.attrib['key'],attr.attrib,attr.text)
                alld[new_parameter.SignalName] = new_parameter
        self.parameters = alld
    def open(self):
        '''
        Open AmdaClient
        '''
        self.client.open_communication_command()
        super(self.__class__, self).open()

    def close(self):
        '''
        Close AmdaClient
        '''
        print('closing')
        self.client.close_communication_command()
        del self.client
        super(self.__class__, self).close()

    @doc_inherit
    def set_parameter(self, param_name, param_value):
        name,key = param_name.split('.')
        path = self.parameters[name].path
        type = self.parameters[name].attr[key].data['type']
        command = 'put {path} {key} {type} {value}'.format( path = path, key = key, type=type, value = param_value)
        self.client.send_command(command)

    def set_caer_sshs(self, path, key, type, value):
        command = 'put {path} {key} {type} {value}'.format( path = path, key = key, type=type, value = value)
        self.client.send_command(command)

    def get_caer_sshs(self, path, key, type):
        command = 'get {path} {key} {type}'.format( path = path, key = key, type=type)
        return self.client.send_command(command)

    def get_param_names(self):
        #CONVENIENCE FUNCTION. IMPLEMENTATION IS NOT REQUIRED
        '''
        Returns names of all the parameters
        '''
        import numpy as np
        kv_list = []
        for name in list(self.parameters.keys()):
            for key, a in list(self.parameters[name].attr.items()):
                if key in ['coarseValue','fineValue']: kv_list.append(name+"."+key)      
        return np.sort(kv_list).tolist()


    @doc_inherit
    def get_parameter(self, param_name):
        if "." in param_name:
            name,key = param_name.split('.')
            path = self.parameters[name].path
            type = self.parameters[name].attr[key].data['type']
            command = 'get {path} {key} {type}'.format( path = path, key = key, type=type)
            value = self.client.send_command(command).strip('\x00'.encode())
            if key in ['coarseValue','fineValue']: value = int(value)
            self.parameters[name].attr[key].value = value
            return value
        else:
            name = param_name
            path = self.parameters[name].path
            kv_list = []
            for key, a in list(self.parameters[name].attr.items()):
                type = a.data['type']
                command = 'get {path} {key} {type}'.format( path = path, key = key, type=type)
                value = int(self.client.send_command(command).strip('\x00'))
                if key in ['coarseValue','fineValue']: value = int(value)
                self.parameters[name].attr[key].value = value
                kv_list.append((name+"."+key,value))
            return kv_list

        

    @doc_inherit
    def reset(self):
        pass




if __name__ == '__main__':
    c = conf_nettcpclient.Configurator(host='gillygaloo.ss.uci.edu')
    c.load_parameter_definitions('/home/eneftci/Projects/code/C/caer/caer-config.xml')



from collections import defaultdict
from contextlib import contextmanager
from pyNCSre.pyST import events
from pyNCSre.pyST import channelEvents
import numpy as np
import errno
import time
import sys
import os
from os import system
from . import globals
from getpass import getuser


def flatten(l):
    return [item for sublist in l for item in sublist]


def is_file_empty(filename):
    return os.stat(filename)[6] == 0


def default_user(user=None):
    if user == None:
        return getuser()
    else:
        return user


def __import_alt__(mod, alt_mod):
    '''
    import mod, if that module throws an ImportError, import alt_mod
    '''
    try:
        return __import__(mod)
    except ImportError:
        return __import__(alt_mod)


def dlist_to_dict(mapping):
    #sort list
    mapping_dict = defaultdict(list)
    func = lambda srctgt: mapping_dict[srctgt[0]].append(srctgt[1])
    list(map(func, mapping))
    return mapping_dict


@contextmanager
def empty_context():
    yield


## {{{ http://code.activestate.com/recipes/576862/ (r1)
"""
doc_inherit decorator

Usage:

class Foo(object):
    def foo(self):
        "Frobber"
        pass

class Bar(Foo):
    @doc_inherit
    def foo(self):
        pass

Now, Bar.foo.__doc__ == Bar().foo.__doc__ == Foo.foo.__doc__ == "Frobber"
"""

from functools import wraps


class DocInherit(object):
    """
    Docstring inheriting method descriptor

    The class itself is also used as a decorator
    """

    def __init__(self, mthd):
        self.mthd = mthd
        self.name = mthd.__name__

    def __get__(self, obj, cls):
        if obj:
            return self.get_with_inst(obj, cls)
        else:
            return self.get_no_inst(cls)

    def get_with_inst(self, obj, cls):

        overridden = getattr(super(cls, obj), self.name, None)

        @wraps(self.mthd, assigned=('__name__', '__module__'))
        def f(*args, **kwargs):
            return self.mthd(obj, *args, **kwargs)

        return self.use_parent_doc(f, overridden)

    def get_no_inst(self, cls):

        for parent in cls.__mro__[1:]:
            overridden = getattr(parent, self.name, None)
            if overridden:
                break

        @wraps(self.mthd, assigned=('__name__', '__module__'))
        def f(*args, **kwargs):
            return self.mthd(*args, **kwargs)

        return self.use_parent_doc(f, overridden)

    def use_parent_doc(self, func, source):
        if source is None:
            func.__doc__ = ''
        else:
            func.__doc__ = source.__doc__
        return func

doc_inherit = DocInherit
## end of http://code.activestate.com/recipes/576862/ }}}
